class PluginsSelection:

    def __init__(self, plugins):
        self.plugins = plugins

    # If there are plugins with same names, this method
    # will return those which have lower versions
    def get_outdated_plugins(self):
        outdated = []
        for p in self.plugins:
            old_plugin = self._get_older_version_of_plugin(p)
            if old_plugin is not None and not old_plugin in outdated:
                outdated.append(old_plugin)
        return outdated

    def _get_older_version_of_plugin(self, plugin):
        namesakes = filter(lambda p: p.get_short_name() == plugin.get_short_name(), self.plugins)
        for older_candidate in namesakes:
            if older_candidate.get_sem_version().less_than(plugin.get_sem_version()):
                return older_candidate
        return None


import os


class Cache:

    def __init__(self, location):
        self.location = location

    @classmethod
    def in_default_location(cls):
        return Cache("/tmp")

    def get(self, dist):

        artifact_location = os.path.join(self.location, dist.get_filename())

        if not os.path.isfile(artifact_location) or "SNAPSHOT" in dist.get_filename():
            dist.download(self.location)
        else:
            print "File %s is already cached." % dist.get_filename()

        return artifact_location


import urllib2
import time


class Utils:

    def __init__(self):
        pass

    __base_download_url = "https://tech.xebialabs.com/download"


    @staticmethod
    def build_product_label(_version):
        return 'xl-deploy' if int(_version.split()[0][0]) >= 4 else 'deployit'

    @staticmethod
    def build_server_zip_name(_version):
        return "{server_dir_name}.zip". \
            format(server_dir_name=Utils.build_server_dir_name(_version))

    @staticmethod
    def build_cli_zip_name(_version):
        return "%s.zip" % Utils.build_cli_dir_name(_version)

    @staticmethod
    def build_cli_zip_url(_version):
        return "%s/". \
            format(server_dir_name=Utils.build_cli_dir_name(_version))

    @staticmethod
    def build_server_dir_name(_version):
        return "{product_label}-{version}-server". \
            format(version=_version, product_label=Utils.build_product_label(_version))

    @staticmethod
    def build_cli_dir_name(_version):
        return "{product_label}-{version}-cli". \
            format(version=_version, product_label=Utils.build_product_label(_version))

    @staticmethod
    def build_plugin_name(_version, _plugin_name):
        return "{plugin_name}-plugin-{version}.jar". \
            format(version=_version, plugin_name=_plugin_name)


    @staticmethod
    def build_url(_version, _xltype, _plugin_short_name=''):
        if _xltype == "server":
            return "{base_url}/{product_label}/{version}/{zip_name}" \
                .format(
                base_url=Utils.__base_download_url,
                version=_version,
                product_label=Utils.build_product_label(_version),
                zip_name=Utils.build_server_zip_name(_version)
            )

        elif _xltype == 'cli':
            return "{base_url}/{product_label}/{version}/{zip_name}" \
                .format(
                base_url=Utils.__base_download_url,
                version=_version,
                product_label=Utils.build_product_label(_version),
                zip_name=Utils.build_cli_zip_name(_version)
            )

        elif _xltype == "plugin":
            return "{base_url}/plugins/{short_plugin_name}-plugin/{version}/{plugin_name}". \
                format(base_url=Utils.__base_download_url, version=_version, short_plugin_name=_plugin_short_name,
                       plugin_name=Utils.build_plugin_name(_version, _plugin_short_name))


        else:
            raise Exception("Unknown artifact type %s for building URL" % _xltype)

    @staticmethod
    def post(method, req):
        # by default method is GET, to make it POST add_data to be called.
        if method == 'POST':
            req.add_data('')

    @staticmethod
    def open_url(url, method, handler=None):

        if handler is None:
            opener = urllib2.build_opener()
        else:
            opener = urllib2.build_opener(handler)

        req = urllib2.Request(url)
        Utils.post(method, req)
        opener.open(req)

        return opener

    @staticmethod
    def auth_and_open_url(url, method, user, password):
        password_mgr = urllib2.HTTPPasswordMgrWithDefaultRealm()
        password_mgr.add_password(None, url, user, password)
        handler = urllib2.HTTPBasicAuthHandler(password_mgr)
        return Utils.open_url(url, method, handler)

    @staticmethod
    def download(url, target, username=None, password=None):
        print "Downloading {url} is being started".format(url=url)

        if username is not None:
            print "Authenticating with username %s" % username

        opener = Utils.auth_and_open_url(url, 'GET', username, password)
        urllib2.install_opener(opener)

        file_name = url.split('/')[-1]
        u = urllib2.urlopen(url)
        target_file = (target + '/' + file_name, target + file_name)[target.endswith('/')]

        with open(target_file, 'wb') as f:
            block_size = 8192

            counter = 0

            while True:
                counter += 1
                if counter % 10 == 0:
                    Utils.print_no_newline(".")
                file_buffer = u.read(block_size)
                if not file_buffer:
                    break

                f.write(file_buffer)

            f.close()

        print ""

        return target_file

    @staticmethod
    def print_no_newline(text):
        import sys

        sys.stdout.write(text)
        sys.stdout.flush()

    @staticmethod
    def wait_until(some_predicate, timeout=20, period=0.25, tick=False):
        must_end = time.time() + timeout
        while time.time() < must_end:
            if some_predicate():
                return True
            if tick:
                print ' ... tick ... '
            time.sleep(period)

        raise Exception("The expected event hasn't happened within %s sec." % timeout)



from xl_helper.actions.Installer import Installer
from xl_helper.actions.Uninstaller import Uninstaller
from xl_helper.artifacts.server.RemoteServerDist import RemoteServerDist
from xl_helper.artifacts.server.LocalServerDist import LocalServerDist


class Engine:
    def __init__(self, path, config=None):
        self.path = path
        self.config = config

    def run(self, op, subject, version, distribution_path=None, upgrade_from_path=None, start=False, plugin_name=None):

        if op == 'install':

            installer = Installer(self.config)

            if subject == 'server':
                dist = RemoteServerDist(version, self.config) if distribution_path is None else LocalServerDist(distribution_path)
                installer.server(
                    dist=dist,
                    target=self.path,
                    upgrade_from_path=upgrade_from_path,
                    start=start
                )
            elif subject == 'cli':
                installer.cli(version, self.path)

            else:
                raise Exception("Unknown subject %s" % subject)
        elif op == 'update':
            print "Update"
        elif op == 'uninstall':
            uninstaller = Uninstaller()

            if plugin_name is None:
                uninstaller.server(version, self.path)
            else:
                uninstaller.plugin(plugin_name, version, self.path)



import re
from xl_helper.artifacts.server.ServerDist import ServerDist


class LocalServerDist(ServerDist):

    name_matcher = re.compile('.*-([0-9\.]+(-SNAPSHOT)?)-server\.zip')

    def __init__(self, path):
        self.path = path
        version_matches = self.name_matcher.match(self.get_filename())
        self.version = version_matches.group(1)

    def get_filename(self):
        return self.path.split('/')[-1]




class SemanticVersion:

    def __init__(self, version):
        version_parts = version.split(".")
        if len(version_parts) < 3:
            raise ValueError("Version '" + version + "' doesn't have a MAJOR.MINOR.PATCH format.")
        self.major = version_parts[0]
        self.minor = version_parts[1]
        self.patch = version_parts[2].split('-')[0]
        self.snapshot = len(version_parts[2].split('-')) > 1

    def less_than(self, other_version):
        if not isinstance(other_version, SemanticVersion):
            raise ValueError("The version '%s' is not of type SemanticVersion." % other_version)

        if self.major != other_version.major:
            return self.major < other_version.major

        if self.minor != other_version.minor:
            return self.minor < other_version.minor

        if self.patch < other_version.patch:
            return True

        return self.snapshot and not other_version.snapshot

    def __str__(self):
        return str(self.__dict__)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


from xl_helper.Utils import Utils

class DownloadsCliDist:

    def __init__(self, version, config):
        self.version = version
        self._downloadsUsername = config.get('downloads', 'username')
        self._downloadsPassword = config.get('downloads', 'password')

    def download(self, target):
        print "Downloading CLI distribution %s" % self.version

        Utils().download(
            url=Utils.build_url(_version=self.version, _xltype="cli"),
            target=target,
            username=self._downloadsUsername,
            password=self._downloadsPassword
        )

        return self

    def get_filename(self):
        return Utils.build_cli_zip_name(self.version)


import os.path
import shutil
from zipfile import ZipFile

from xl_helper.Utils import Utils
from Server import Server
from xl_helper.artifacts.Plugin import Plugin
from xl_helper.artifacts.Cache import Cache
from xl_helper.artifacts.PluginsSelection import PluginsSelection
from distutils import dir_util
from xl_helper.artifacts.cli.DownloadsCliDist import DownloadsCliDist
from xl_helper.artifacts.server.LocalServerDist import LocalServerDist


class Installer:

    cache = Cache.in_default_location()

    def __init__(self, config):
        self.license_location = os.path.expanduser(config.get('license', 'location'))
        self.license_location_3x = os.path.expanduser(config.get('license', 'location_3x'))
        self.config = config

    def server(self, dist, target, upgrade_from_path=None, start=False):

        source = dist.path if isinstance(dist, LocalServerDist) else self.cache.get(dist)

        ZipFile(source, 'r').extractall(target)

        # Copy available plugins
        server_dir = os.path.join(target, '.'.join(dist.get_filename().split('.')[:-1]))

        print "Server has been installed to %s" % server_dir

        for root, dirs, files in os.walk(os.path.join(server_dir, 'available-plugins')):
            for file_ in files:
                print "Copying available plugin %s" % file_
                if file_.endswith('.jar'):
                    shutil.copy(os.path.join(root, file_), os.path.join(server_dir, 'plugins', file_))

        proper_license = self.license_location if (dist.version.startswith('4') or dist.version == 'SNAPSHOT') else self.license_location_3x

        print "Copying license from %s" % proper_license
        shutil.copy(proper_license, os.path.join(server_dir, 'conf', 'deployit-license.lic'))

        for dirpath, d, file_names in os.walk(target):
            for filename in file_names:
                if filename.endswith(".sh"):
                    os.chmod(dirpath + '/' + filename, 0750)

        if upgrade_from_path is not None:
            print "Copying files from old installation at %s" % upgrade_from_path
            if os.path.isdir(os.path.join(upgrade_from_path, 'repository')):
                dir_util.copy_tree(os.path.join(upgrade_from_path, 'repository'), os.path.join(server_dir, 'repository'))
            dir_util.copy_tree(os.path.join(upgrade_from_path, 'plugins'), os.path.join(server_dir, 'plugins'))
            dir_util.copy_tree(os.path.join(upgrade_from_path, 'conf'), os.path.join(server_dir, 'conf'))
            dir_util.copy_tree(os.path.join(upgrade_from_path, 'ext'), os.path.join(server_dir, 'ext'))
            self._remove_old_plugins(server_dir)

        if start:
            Server.from_config(config=self.config, home=server_dir).start()

        return server_dir

    def _remove_old_plugins(self, server_location):
        plugins_path = os.path.join(server_location, 'plugins')
        for root, dirs, files in os.walk(plugins_path):
            plugins_selection = PluginsSelection(map(Plugin, filter(Plugin.is_plugin, files)))
            for op in plugins_selection.get_outdated_plugins():
                os.remove(os.path.join(plugins_path, op.filename))

    def plugin(self, name, version, server_location):
        plugins_path = os.path.join(server_location, 'plugins')

        print "Installing plugin %s into " % plugins_path

        u = Utils.build_url(_version=version, _xltype="plugin", _plugin_short_name=name)

        Utils().download(
            url=u,
            target=plugins_path,
            username=self.config.get('downloads', 'username'),
            password=self.config.get('downloads', 'password')
        )

        server = Server.from_config(config=self.config, home=server_location)

        if server.is_running():
            server.restart()

    def cli(self, version, path):
        ZipFile(self.cache.get(DownloadsCliDist(version, self.config)), 'r').extractall(path)
        return os.path.join(path, Utils.build_cli_dir_name(version))







import os
import shutil
import urllib2
from urlparse import urljoin
from xl_helper.FileUtils import FileUtils
from subprocess import Popen, PIPE, STDOUT
from xl_helper.Utils import Utils


class Server:

    @staticmethod
    def from_config(config, home):
        return Server(
            home=home,
            username=config.get('deployit', 'username'),
            password=config.get('deployit', 'password'),
            url="{protocol}://{host}:{port}/deployit/".format(
                protocol=config.get('deployit', 'protocol'),
                host=config.get('deployit', 'host'),
                port=config.get('deployit', 'port')
            )
        )

    def __init__(self, home, username, password, url):
        self.home = home
        self.username = username
        self.password = password
        self.url = url + '/' if url[-1] != '/' else url

    def is_stopped(self):
        return not self._ping()

    def is_running(self):
        return self._ping()

    def restart(self):
        self.stop()
        self.start()

    def start(self):
        if self.is_stopped():
            server_sh_file = "{}/bin/server.sh".format(self.home)

            if not os.path.isdir("{path}/repository".format(path=self.home)):
                print "Initializing XL Deploy with default configuration."
                shutil.copy(FileUtils.to_absolute_path("resources/deployit.conf"), os.path.join(self.home, "conf"))
                p = Popen([server_sh_file, '-setup', '-reinitialize'], bufsize=0, stdin=PIPE, stdout=PIPE, stderr=STDOUT)
                p.stdin.write("yes\n")
                p.stdin.flush()
                p.communicate()

            if not os.path.isfile(server_sh_file):
                print "Server cannot be started as file {} hasn't been found".format(server_sh_file)
            else:
                os.system(server_sh_file)
        else:
            print "Server is already started"

    def stop(self):
        if self.is_running():
            stop_url = self.url + 'server/shutdown'
            opener = Utils.auth_and_open_url(stop_url, 'POST', self.username, self.password)
            print "Stopping the server"
            urllib2.install_opener(opener)

    def _ping(self):
        try:
            ping_url = self.url + 'server/info'
            print "Pinging %s " % ping_url
            opener = Utils.auth_and_open_url(ping_url, 'GET', self.username, self.password)
            urllib2.install_opener(opener)
            return True
        except urllib2.URLError:
            return False



import re

from xl_helper.artifacts.SemanticVersion import SemanticVersion


class Plugin:

    plugin_name_matcher = re.compile('.*-plugin-([0-9\.]+(-SNAPSHOT)?)\.jar')

    @staticmethod
    def is_plugin(path):
        return Plugin.plugin_name_matcher.match(path) is not None

    def __init__(self, filename):
        self.filename = filename

    def get_short_name(self):
        return self.filename.split("-plugin-")[0]

    def get_version(self):
        version_matches = self.plugin_name_matcher.match(self.filename)
        return version_matches.group(1)

    def get_sem_version(self):
        return SemanticVersion(self.get_version())

    def has_same_name_as(self, another_plugin):
        return self.get_short_name() == another_plugin.get_short_name()

    def __str__(self):
        return str(self.__dict__)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


from argparse import ArgumentParser
import os


class Argh:

    @staticmethod
    def createParser():
        parser = ArgumentParser(description='XL helper')
        parser.add_argument('action', choices=['install', 'update', 'uninstall'], type=str, help='What you want to do.')
        parser.add_argument('subject', choices=['server', 'plugin', 'cli'], type=str, help='What will be affected.')
        parser.add_argument('plugin', metavar='PLUGIN_ID', nargs='?', type=str, help='Plugin id.', default=None)
        parser.add_argument('--version', metavar='VER', type=str, help='Version. No version means latest snapshot.', required=False, default='SNAPSHOT')
        parser.add_argument('--dist', metavar='DIST', type=str, help='Distribution path. Instead of downloading, use the local distribution zip/jar.', required=False, default=None)
        parser.add_argument('--path', metavar='PATH', type=str, help='Path to the server installation.', required=False, default=os.getcwd())
        parser.add_argument('--upgrade-from-path', metavar='PATH', type=str, help='Path to the server installation from which it should upgrade.', required=False, default=None)

        parser.add_argument('-start', type=bool, help='Make sure that server is running after the action.', nargs='?', default=False, const=True)
        return parser



#!/usr/bin/env python

from setuptools import setup, find_packages

setup(name='xl-helper',
      version='1.0.3',
      description='XL Deploy helper',
      long_description='This tool helps with installation and upgrade of XL Deploy and plugins',
      author='Mike Kotsur',
      author_email='mkotsur@xebialabs.com',
      url='http://xebialabs.com/',
      packages=find_packages(where=".", exclude=["tests*"]),
      install_requires=['jenkinsapi', 'argparse'],
      scripts=['bin/xl-helper']
)



import os.path
import shutil

from xl_helper.Utils import Utils


class Uninstaller:

    def server(self, version, location):
        server_dir = os.path.join(location, Utils.build_server_dir_name(version))
        if os.path.isdir(server_dir):
            shutil.rmtree(server_dir)
        else:
            print "The server {} hasn\'t been removed as it doesn\'t exist".format(server_dir)

        zip_path = os.path.join(location, Utils.build_server_zip_name(version))
        if os.path.isfile(zip_path):
            os.remove(zip_path)
        else:
            print "Zip file {} hasn\'t been removed as it doesn\'t exist".format(zip_path)

    def plugin(self, name, version, location):
        server_dir = os.path.join(location, Utils.build_server_dir_name(version))
        plugin_path = "{path}/plugins/{name}".format(path=server_dir, name=Utils.build_plugin_name(version, name))

        if os.path.isfile(plugin_path):
            os.remove(plugin_path)
        else:
            print "Plugin {} hasn\'t been removed as it doesn\'t exist".format(plugin_path)



from xl_helper.Utils import Utils
from xl_helper.artifacts.server.ServerDist import ServerDist
from jenkinsapi.jenkins import Jenkins


class RemoteServerDist(ServerDist):

    def __init__(self, version, config):
        self.config = config
        self.version = version
        self._jobName = self.config.get('jenkins', 'server-job')
        self._downloadsUsername = self.config.get('downloads', 'username')
        self._downloadsPassword = self.config.get('downloads', 'password')
        self._jenkins = Jenkins(
            self.config.get('jenkins', 'url'),
            self.config.get('jenkins', 'username'),
            self.config.get('jenkins', 'password'),
            lazy=True
        )

    def download(self, target):
        print "Downloading: %s" % self.get_filename()
        username = self._jenkins.username if self.version == 'SNAPSHOT' else self._downloadsUsername
        password = self._jenkins.password if self.version == 'SNAPSHOT' else self._downloadsPassword
        return Utils().download(url=self._get_url(), target=target, username=username, password=password)

    def _get_url(self):

        if self.version == 'SNAPSHOT':
            job = self._jenkins.get_job(self._jobName)
            artifacts = job.get_last_good_build().get_artifact_dict().values()
            server_artifact = [a for a in artifacts if a.filename.endswith('-server.zip')].pop()
            return server_artifact.url
        else:
            return Utils.build_url(_version=self.version, _xltype="server")

    def get_filename(self):
        return self._get_url().split('/')[-1]



import inspect
import os
import shutil


class FileUtils:

    def __init__(self):
        pass

    @staticmethod
    def delete_dirs(*paths):
        for path in paths:
            if os.path.isdir(path):
                shutil.rmtree(path)

    @staticmethod
    def delete_contents(location):
        for root, dirs, files in os.walk(location):
            for d in dirs:
                FileUtils.delete_dirs(os.path.join(root, d))
            for f in files:
                shutil.rmtree(os.path.join(root, f))

    @staticmethod
    def copy_contents(src, dst):
        if not os.path.isdir(src):
            return

        if not os.path.isdir(dst):
            os.mkdir(dst)

        for root, dirs, files in os.walk(src):
            for d in dirs:
                shutil.copytree(os.path.join(root, d), dst)
            for f in files:
                shutil.copyfile(os.path.join(root, f), os.path.join(dst, f))

    @staticmethod
    def to_absolute_path(relative_path):
        return os.path.normpath(os.path.join(os.path.dirname(__file__ ), '..', relative_path))

    @staticmethod
    def proj_folder():
        cur_script_folder = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
        return os.path.split(cur_script_folder)[0]

    @staticmethod
    def conf_folder():
        return "{proj_dir}/conf".format(proj_dir=FileUtils.proj_folder())



import os.path
import shutil

from Server import Server
from Installer import Installer
from xl_helper.FileUtils import FileUtils


# class Updater:
#
#     def __init__(self, config):
#         self.config = config
#         self.installer = Installer(config)
#
#     def server(self, dist, location, start=False):
#         backup_location = location + '_old'
#         FileUtils.delete_dirs(backup_location)
#         shutil.copytree(location, backup_location)
#         FileUtils.delete_contents(location)
#
#         new_version_dir = self.installer.server(dist, os.path.join(location, os.path.pardir), start)
#
#         print "new_version_dir: " + new_version_dir
#
#         FileUtils.copy_contents(new_version_dir, location)
#
#         FileUtils.copy_contents(os.path.join(backup_location, 'repository'), os.path.join(location, 'repository'))
#         FileUtils.copy_contents(os.path.join(backup_location, 'plugins'), os.path.join(location, 'plugins'))
#         FileUtils.copy_contents(os.path.join(backup_location, 'conf'), os.path.join(location, 'conf'))
#
#         FileUtils.delete_dirs(backup_location, new_version_dir)
#         pass
